#ifdef _WIN32
#  define WINDOWS_LEAN_AND_MEAN
#  define NOMINMAX
#  include <windows.h>
#pragma warning(disable:4996)
#endif

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, GL
#include <GL/glew.h>

#if defined(__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <cuda_runtime_api.h>
#include <cuda_gl_interop.h>
#include <cutil_inline.h>
#include <cutil_gl_error.h>
#include <cudpp/cudpp.h>

#include <imageApprox_kernel.cu>

#define TEST1
/*** uncomment the following line to run test2 ***/
//#define TEST2

/*** uncomment the following line will cause the CPU solution
 *** to output color-coded tiles, for the convenience of debugging ***/
//#define DEBUG

#ifdef TEST1
char *image_filename = "lena512.pgm";
const float initial_error = 0.01f;
const unsigned int initial_tilesize = 32;
#endif

#ifdef TEST2
char *image_filename = "earth4096.pgm";
const float initial_error = 0.005f;
const unsigned int initial_tilesize = 128;
#endif

unsigned int image_width = 0, image_height = 0;
int cpu_ntiles = 0;

// declaration, forward
void runTest(int argc, char** argv);

int main(int argc, char** argv)
{
    runTest(argc, argv);

    return EXIT_SUCCESS;
}

void fillImageTile(float* input, float* output, unsigned int x, unsigned int y, unsigned int tilesize,
						unsigned int image_width, unsigned int image_height)
{
	unsigned int i;
	unsigned int j;
	float s, t;
	
	/* extract the four corner pixels of the tile */
	float p0 = input[y*image_width+x];
	float p1 = input[y*image_width+x+(tilesize-1)];
	float p2 = input[(y+tilesize-1)*image_width+x];
	float p3 = input[(y+tilesize-1)*image_width+x+(tilesize-1)];

	float colorcode = rand() / (float) RAND_MAX;

	for (j = 0; j < tilesize; j ++)
	{
		t = j / (float) (tilesize-1);
		for (i = 0; i < tilesize; i ++)
		{
			s = i / (float) (tilesize-1);

			// linearly interpolate from the four corner pixels
			output[(y+j)*image_width+(x+i)] = (1.0-s)*(1.0-t)*p0 + s*(1.0-t)*p1 + (1.0-s)*t*p2 + s*t*p3;

#ifdef DEBUG
			// output random color-coding
			output[(y+j)*image_width+(x+i)] = colorcode;
#endif
		}
	}

	// count image tiles
	cpu_ntiles ++;
}

void processImageTile(float* input, float* output, unsigned int x, unsigned int y, unsigned int tilesize,
						unsigned int image_width, unsigned int image_height, float error)
{
	// if tile is small enough, stop recursion
	if (tilesize <= 2)
	{
		fillImageTile(input, output, x, y, tilesize, image_width, image_height);
		return;
	}

	// extract four corner values
	float p0 = input[y*image_width+x];
	float p1 = input[y*image_width+(x+tilesize-1)];
	float p2 = input[(y+tilesize-1)*image_width+x];
	float p3 = input[(y+tilesize-1)*image_width+(x+tilesize-1)];

	// run error test
	if ((fabs(p0-p1)/(p0+p1) <= error) &&
		(fabs(p0-p2)/(p0+p2) <= error) &&
		(fabs(p0-p3)/(p0+p3) <= error) &&
		(fabs(p1-p2)/(p1+p2) <= error) &&
		(fabs(p1-p3)/(p1+p3) <= error) &&
		(fabs(p2-p3)/(p2+p3) <= error) )
	{
		// passed error test, fill tile's internal pixel values via linear interpolation
		fillImageTile(input, output, x, y, tilesize, image_width, image_height);
		return;
	}
	else
	{
		// failed error test, subdivide into 4 sub-tiles and recurse on each sub-tile
		// double the error threshold, meaning we progressively tolerate higher errors for smaller tiles
		processImageTile(input, output, x, y, tilesize/2, image_width, image_height, error*2.0);
		processImageTile(input, output, x+tilesize/2, y, tilesize/2, image_width, image_height, error*2.0);
		processImageTile(input, output, x, y+tilesize/2, tilesize/2, image_width, image_height, error*2.0);
		processImageTile(input, output, x+tilesize/2, y+tilesize/2, tilesize/2, image_width, image_height, error*2.0);
	}
}

void computeGold(float* input, float* output)
{

	clock_t timer = clock();
	unsigned int i, j;
	unsigned int x, y;
	unsigned int tilesize = initial_tilesize;
	float error = initial_error;

	for (j = 0; j < image_height/tilesize; j++) 
	{
		y = j * tilesize;
		for (i = 0; i < image_width/tilesize; i++)
		{
			x = i * tilesize;
			processImageTile(input, output, x, y, tilesize, image_width, image_height, error);
		}
	}

	timer = clock() - timer;
	printf("CPU solution time: %d ms.\n", (int) timer);
	printf("Number of tiles: %d.\n", cpu_ntiles);
	printf("-------------------------------------------\n");
}

void computeOnDevice(float* h_input, float* h_output)
{

	/* Allocate device memory, transfer data to device */
	/* this part should not be couted into timing */
	
	/* the following code sets up a CUDA texture for the intput data */
	cudaChannelFormatDesc channelDesc =
		cudaCreateChannelDesc(32, 0, 0, 0, cudaChannelFormatKindFloat);	// 1 float per pixel
	
	cudaArray* d_tex_cuarray;
	cutilSafeCall( cudaMallocArray(&d_tex_cuarray,
								   &channelDesc,
								   image_width,
								   image_height));	// allocate cuArray

	cutilSafeCall( cudaMemcpyToArray( d_tex_cuarray, 0, 0,
									  h_input, image_width*image_height*sizeof(float),
									  cudaMemcpyHostToDevice));	// transfer data to cuArray

	// setup texture mode. 'tex' is defined in imageApprox_kernel.cu
	tex.addressMode[0] = cudaAddressModeWrap;
	tex.addressMode[1] = cudaAddressModeWrap;
	tex.filterMode = cudaFilterModeLinear;
	tex.normalized = false;	// un-normalized texture indexing, so that you can access
							// the texture with an index in the range [0,image_width)x[0,image_height)
							// just like how you index into CPU memory
	cutilSafeCall( cudaBindTextureToArray( tex, d_tex_cuarray, channelDesc)); // bind cuArray to texture

	unsigned int timer;
	cutCreateTimer(&timer);
	cutStartTimer(timer);

	/***********************/
	/****** YOUR CODE ******/
	/***********************/

	cudaThreadSynchronize();
	cutStopTimer(timer);

	printf("GPU solution time: %d ms.\n", (int) timer);
	printf("Number of tiles: %d.\n", 0);
	printf("-------------------------------------------\n");
	cutDeleteTimer(timer);

	/* Transfer data back to host */
	/* this part should not be couted into timing */

}

void runTest(int argc, char** argv)
{
	// init cuda
    cutilDeviceInit(argc, argv);

    // load image from disk
    float* h_input = NULL;
    char* image_path = cutFindFilePath(image_filename, argv[0]);
    char output_filename[1024];
	if (image_path == 0) {
        printf("Unable to source file file %s\n", image_filename);
        exit(EXIT_FAILURE);
    }
    cutilCheckError( cutLoadPGMf(image_path, &h_input, &image_width, &image_height));
    printf("Source image loaded '%s', %d x %d pixels\n", image_filename, image_width, image_height);

	float *h_output = new float[image_width*image_height];

	computeGold(h_input, h_output);

	// wirte image to disk file
    strcpy(output_filename, image_path);
    strcpy(output_filename + strlen(image_path) - 4, "_output_gold.pgm");
	cutilCheckError( cutSavePGMf( output_filename, h_output, image_width, image_height));

	memset(h_output, 0, sizeof(float)*image_width*image_height);

	computeOnDevice(h_input, h_output);

    strcpy(output_filename, image_path);
    strcpy(output_filename + strlen(image_path) - 4, "_output_gpu.pgm");
	cutilCheckError( cutSavePGMf( output_filename, h_output, image_width, image_height));
}

